/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2008 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: hp.c,v 1.84 2008/08/05 13:35:48 zlb Exp $
 *
 * Support functions for hp-adaptivity */

#include "phg.h"
#include <string.h>

HP_TYPE *
phgHPNew_(GRID *g, HP_INFO *info, int order, const char *file, int line)
/* allocates and returns a new HP_TYPE object.
 * 'order' is the initial order to use (-1 ==> no initial order) */
{
    HP_TYPE *hp;
    size_t len;
    INT i;

    hp = phgCalloc(1, sizeof(*hp));
    hp->info = info;
    hp->srcfile = strdup(file);
    hp->srcline = line;
    hp->g = g;

    hp->edge_order = phgAlloc(g->nedge * sizeof(*hp->edge_order));
    hp->face_order = phgAlloc(g->nface * sizeof(*hp->face_order));
    hp->elem_order = phgAlloc(g->nelem * sizeof(*hp->elem_order));

    if (order < 0) {
	memset(hp->edge_order, 0, g->nedge * sizeof(*hp->edge_order));
	memset(hp->face_order, 0, g->nface * sizeof(*hp->face_order));
	memset(hp->elem_order, 0, g->nelem * sizeof(*hp->elem_order));
    }
    else {
	INT a[3], nedge, nface;

	if (order < info->min_order)
	    order = info->min_order;
	else if (order > info->max_order)
	    order = info->max_order;
	for (i = 0; i < g->nedge; i++)
	    hp->edge_order[i] = order;
	for (i = 0; i < g->nface; i++)
	    hp->face_order[i] = order;
	for (i = 0; i < g->nelem; i++)
	    hp->elem_order[i] = order;
	phgHPSetupDataPointers(hp);

	/* update counters */
	hp->max_order = hp->min_order = order;

	nedge = 0;
	for (i = 0; i < g->nedge; i++) {
	    if ((g->types_edge[i] & OWNER))
		nedge += hp->edge_index[i + 1] - hp->edge_index[i];
	}

	nface = 0;
	for (i = 0; i < g->nface; i++) {
	    if ((g->types_face[i] & OWNER))
		nface += hp->face_index[i + 1] - hp->face_index[i];
	}

	a[0] = nedge;
	a[1] = nface;
	a[2] = hp->elem_index[g->nelem];
#if USE_MPI
	if (g->nprocs > 1) {
	    INT b[3];
	    MPI_Allreduce(a, b, 3, PHG_MPI_INT, MPI_SUM, g->comm);
	    memcpy(a, b, sizeof(b));
	}
#endif
	hp->edge_data_count_global = a[0];
	hp->face_data_count_global = a[1];
	hp->elem_data_count_global = a[2];
    }

    /* attach hp to g->hp */
    len = 0;
    if (g->hp != NULL) {
	HP_TYPE **p;
	for (p = g->hp; *p != NULL; p++, len++);
    }
    g->hp = phgRealloc(g->hp, (len + 2) * sizeof(*(g->hp)));
    g->hp[len] = hp;
    g->hp[len + 1] = NULL;

    return hp;
}

void
phgHPSetupDataPointers_(HP_TYPE *hp, INT nedge, INT nface, INT nelem,
			BOOLEAN use_types, BOOLEAN update_counters)
 /* initializes data pointers, i.e., sets up {edge,face,elem}_index arrays,
  * if 'update_counters' == TRUE than also updates global data counters.
  * if 'use_types' == FALSE than don't use g->types_xxxx[] */
{
    GRID *g = hp->g;
    INT i, k, a[3];

    assert(!update_counters || use_types);

    assert(nedge == 0 || hp->edge_order != NULL);
    assert(nface == 0 || hp->face_order != NULL);
    assert(nelem == 0 || hp->elem_order != NULL);

    if (hp->edge_index != NULL)
	phgFree(hp->edge_index);
    hp->edge_index = phgAlloc((nedge + 1) * sizeof(*hp->edge_index));
    hp->edge_index[0] = a[0] = 0;
    for (i = 0; i < nedge; i++) {
	if (use_types && g->types_edge[i] == UNREFERENCED) {
	    hp->edge_index[i + 1] = hp->edge_index[i];
	    continue;
	}
	k = hp->info->types[hp->edge_order[i]]->np_edge;
	hp->edge_index[i + 1] = hp->edge_index[i] + k;
	if (update_counters && (g->types_edge[i] & OWNER))
	    a[0] += k;
    }

    if (hp->face_index != NULL)
	phgFree(hp->face_index);
    hp->face_index = phgAlloc((nface + 1) * sizeof(*hp->face_index));
    hp->face_index[0] = a[1] = 0;
    for (i = 0; i < nface; i++) {
	if (use_types && g->types_face[i] == UNREFERENCED) {
	    hp->face_index[i + 1] = hp->face_index[i];
	    continue;
	}
	k = hp->info->types[hp->face_order[i]]->np_face;
	hp->face_index[i + 1] = hp->face_index[i] + k;
	if (update_counters && (g->types_face[i] & OWNER))
	    a[1] += k;
    }

    if (hp->elem_index != NULL)
	phgFree(hp->elem_index);
    hp->elem_index = phgAlloc((nelem + 1) * sizeof(*hp->elem_index));
    hp->elem_index[0] = a[2] = 0;
    for (i = 0; i < nelem; i++) {
	if (use_types && g->types_elem[i] == UNREFERENCED) {
	    hp->elem_index[i + 1] = hp->elem_index[i];
	    continue;
	}
	k = hp->info->types[hp->elem_order[i]]->np_elem;
	hp->elem_index[i + 1] = hp->elem_index[i] + k;
	assert(use_types == FALSE || (g->types_elem[i] & OWNER));
	if (update_counters)
	    a[2] += k;
    }

    if (update_counters) {
#if USE_MPI
	INT b[3];
	MPI_Allreduce(a, b, 3, PHG_MPI_INT, MPI_SUM, g->comm);
	hp->edge_data_count_global = b[0];
	hp->face_data_count_global = b[1];
	hp->elem_data_count_global = b[2];

	a[0] = -hp->min_order;
	a[1] = hp->max_order;
	MPI_Allreduce(a, b, 2, PHG_MPI_INT, MPI_MAX, g->comm);
	hp->min_order = -b[0];
	hp->max_order = b[1];
#else	/* USE_MPI */
	hp->edge_data_count_global = a[0];
	hp->face_data_count_global = a[1];
	hp->elem_data_count_global = a[2];
#endif	/* USE_MPI */
    }

    return;
}

#if 0
# define CMP_ORDER(a,b)	((a) > (b))	/* order = max(orders) (don't use!) */
#else
# define CMP_ORDER(a,b)	((a) < (b))	/* order = min(orders) */
#endif

void
phgHPSetupOrders(HP_TYPE *hp)
/* assumes hp->{info,elem_order} are valid and updates other members */
{
    GRID *g = hp->g;
    INT i;
    SIMPLEX *e;
    int ii, order;

    assert(g->nelem == 0 || hp->elem_order != NULL);

    if (hp->edge_order != NULL)
	phgFree(hp->edge_order);
    hp->edge_order = phgCalloc(g->nedge, sizeof(*hp->edge_order));

    if (hp->face_order != NULL)
	phgFree(hp->face_order);
    hp->face_order = phgCalloc(g->nface, sizeof(*hp->face_order));

    /* Note: global min/max order will be computed by phgHPSetupDataPointers */
    hp->min_order = hp->info->max_order;
    hp->max_order = hp->info->min_order;

    for (i = 0; i < g->nedge; i++)
	hp->edge_order[i] = (CMP_ORDER(255, 0) ? 0 : 255);
    for (i = 0; i < g->nface; i++)
	hp->face_order[i] = (CMP_ORDER(255, 0) ? 0 : 255);

    ForAllElements(g, e) {
	order = hp->elem_order[e->index];
	assert(order >= hp->info->min_order && order <= hp->info->max_order);
	if (hp->min_order > order)
	    hp->min_order = order;
	if (hp->max_order < order)
	    hp->max_order = order;
	for (ii = 0; ii < NEdge; ii++) {
	    i = e->edges[ii];
	    if (!CMP_ORDER(hp->edge_order[i], order))
		hp->edge_order[i] = order;
	}
	for (ii = 0; ii < NFace; ii++) {
	    i = e->faces[ii];
	    if (!CMP_ORDER(hp->face_order[i], order))
		hp->face_order[i] = order;
	}
    }

#if USE_MPI
    if (g->nprocs > 1) {
	/* Note: a dummy MAP is used to build communication pattern */
	static DOF_TYPE type = {DofCache,
	    "dummy", NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    NULL, NULL, NULL, FE_None,
	    TRUE, FALSE, -1, NEdge + NFace, 0, -1, -1,
	    0, 1, 1, 0
	};
	MAP *map;
	BYTE *sbuff, *rbuff;
	INT *V2Dmap;
	DOF *utmp;
	INT k;

	utmp = phgDofNew(g, &type, 1, "dummy", DofNoData);

	map = phgMapCreate(utmp, NULL);

	/* build V2Dmap (reverse of D2Lmap * L2Vmap) */
	V2Dmap = phgAlloc(map->localsize * sizeof(*V2Dmap));
	for (i = 0; i < DofGetDataCount(utmp); i++) {
	    if ((k = phgMapD2L(map, 0, i)) < 0)
		continue;
	    V2Dmap[map->L2Vmap == NULL ? k : map->L2Vmap[k]] = i;
	}

	map->cinfo = phgMapCreateCommInfo(
				map->comm,
				map->nprocs,
				map->rank,
				map->nlocal,
				map->localsize,
				map->O2Gmap,
				map->ordering,
				map->partition);
	sbuff = phgAlloc((map->cinfo->ssize + map->cinfo->rsize)
				* sizeof(*sbuff));
	rbuff = sbuff + map->cinfo->ssize;

	/* copy orders on unowned edges/faces to rbuff */
	for (i = 0; i < map->cinfo->rsize; i++) {
	    k = map->nlocal + map->cinfo->rind[i];
	    assert(k < map->localsize);
	    k = V2Dmap[k];
	    if (k < g->nedge) {
		/* edge data */
		rbuff[i] = hp->edge_order[k];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		rbuff[i] = hp->face_order[k];
	    }
	}

	/* send orders on unowned edges/faces to their owners */
	if (map->cinfo->comm == MPI_COMM_NULL)
	    goto skip_comm;
	phgAlltoallv(rbuff, map->cinfo->rcnts, map->cinfo->rdsps, PHG_MPI_BYTE,
		     sbuff, map->cinfo->scnts, map->cinfo->sdsps, PHG_MPI_BYTE,
		     map->cinfo->comm);

	/* compare with incoming orders */
	for (i = 0; i < map->cinfo->ssize; i++) {
	    k = map->cinfo->sind[i];
	    assert(k < map->nlocal);
	    k = V2Dmap[k];
	    if (k < g->nedge) {
		/* edge data */
		if (!CMP_ORDER(hp->edge_order[k], sbuff[i]))
		    hp->edge_order[k] = sbuff[i];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		if (!CMP_ORDER(hp->face_order[k], sbuff[i]))
		    hp->face_order[k] = sbuff[i];
	    }
	}

	/* update incoming orders */
	for (i = 0; i < map->cinfo->ssize; i++) {
	    k = map->cinfo->sind[i];
	    assert(k < map->nlocal);
	    k = V2Dmap[k];
	    if (k < g->nedge) {
		/* edge data */
		sbuff[i] = hp->edge_order[k];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		sbuff[i] = hp->face_order[k];
	    }
	}

	/* send back incoming orders */
	phgAlltoallv(sbuff, map->cinfo->scnts, map->cinfo->sdsps, PHG_MPI_BYTE,
		     rbuff, map->cinfo->rcnts, map->cinfo->rdsps, PHG_MPI_BYTE,
		     map->cinfo->comm);

	/* update orders on unowned edges/faces */
	for (i = 0; i < map->cinfo->rsize; i++) {
	    k = map->nlocal + map->cinfo->rind[i];
	    assert(k < map->localsize);
	    k = V2Dmap[k];
	    if (k < g->nedge) {
		/* edge data */
		hp->edge_order[k] = rbuff[i];
	    }
	    else {
		k -= g->nedge;
		assert(k < g->nface);
		hp->face_order[k] = rbuff[i];
	    }
	}

skip_comm:
	phgFree(V2Dmap);
	phgFree(sbuff);
	phgMapDestroy(&map);
	phgDofFree(&utmp);
    }
#endif	/* USE_MPI */

    phgHPSetupDataPointers(hp);
}

static DOF *u0;

static void
tmp_func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
{
    Unused(bno);
    phgDofEval(u0, e, lambda, values);
}

static void
update_dof(DOF *u, HP_TYPE *hp, BOOLEAN interpolate)
/* update 'u' according to new hp settings */
{
    GRID *g = u->g;
    INT i, k;
    FLOAT *data, *pdata, *pdata0;
    DOF *utmp;
    DOF_TYPE *oldtype, *newtype;
    int ii, oldcount, newcount;
    DOF_USER_FUNC func0;
    DOF_USER_FUNC_LAMBDA func1;
    VEF_MAP *vef;
    BOOLEAN has_userfunc;

    /* set up data */
    pdata = data = phgCalloc(g->nvert * u->count_vert +
			      u->dim * (hp->edge_index[g->nedge] +
					hp->face_index[g->nface] +
					hp->elem_index[g->nelem]),
			     sizeof(*data));

    if (!interpolate)
	goto skip;

    if (u->userfunc == DofLambdaFunction) {
	assert(u->userfunc_lambda != NULL);
	func0 = NULL;
	func1 = u->userfunc_lambda;
	has_userfunc = TRUE;
    }
    else if (u->userfunc != DofNoAction && u->userfunc != DofInterpolation) {
	func0 = u->userfunc;
	func1 = NULL;
	has_userfunc = TRUE;
    }
    else {
	func0 = NULL;
	func1 = tmp_func;
	has_userfunc = FALSE;
    }

    utmp = phgHPDofNew(g, hp, u->dim, "tmp", DofNoData);
    utmp->data = utmp->data_vert = data;
    utmp->data_edge = utmp->data_vert + g->nvert * u->count_vert;
    utmp->data_face = utmp->data_edge + hp->edge_index[g->nedge] * u->dim;
    utmp->data_elem = utmp->data_face + hp->face_index[g->nface] * u->dim;

    newtype = hp->info->types[hp->max_order];

    /* copy vertex data */
    if ((k = DofGetVertexDataCount(u)) > 0) {
	memcpy(pdata, u->data, k * sizeof(*data));
	pdata += k;
    }

    vef = phgDofSetupVEFMap(g, u, -1);
    u0 = u;

    /* set up edge data */
    for (i = 0; i < g->nedge; i++) {
	newcount = (hp->edge_index[i + 1] - hp->edge_index[i]) * u->dim;
	if (newcount == 0)
	    continue;
	oldcount = DofGetDataCountOnEdge(u, i);
	if (oldcount == 0) {
	    pdata += newcount;
	    continue;
	}
	if (DofIsHP(u) && u->hp->info == hp->info &&
	    oldcount <= newcount && (!has_userfunc || oldcount == newcount)) {
	    if (oldcount > 0)
		memcpy(pdata, DofEdgeData(u, i), oldcount * sizeof(*pdata));
	}
	else {
	    newtype->InitFunc(utmp, vef->Emap[i], EDGE, vef->Eind[i],
				func0, func1, NULL, pdata, NULL);
	}
	pdata += newcount;
    }

    /* set up face data */
    for (i = 0; i < g->nface; i++) {
	newcount = (hp->face_index[i + 1] - hp->face_index[i]) * u->dim;
	if (newcount == 0)
	    continue;
	oldcount = DofGetDataCountOnFace(u, i);
	if (oldcount == 0) {
	    pdata += newcount;
	    continue;
	}
	if (DofIsHP(u) || u->hp->info != hp->info ||
	    oldcount > newcount || (has_userfunc && oldcount != newcount)) {
	    newtype->InitFunc(utmp, vef->Fmap[i], FACE, vef->Find[i],
				func0, func1, NULL, pdata, NULL);
	    pdata += newcount;
	    continue;
	}
	newtype = hp->info->types[hp->face_order[i]];
	oldtype = u->hp->info->types[u->hp->face_order[i]];
	pdata0 = DofFaceData(u, i);
	if (newtype != oldtype && hp->info->types == DOF_HCn) {
	    /* copy edge-based DOF */
	    int nnew = DOF_HBn[newtype->order]->np_edge * u->dim;
	    int nold = DOF_HBn[oldtype->order]->np_edge * u->dim;
	    assert(oldtype->order >= 2 && newtype->order >= 2);
	    for (ii = 0; ii < 3; ii++) {
		memcpy(pdata, pdata0, nold * sizeof(*pdata));
		pdata0 += nold;
		oldcount -= nold;
		pdata += nnew;
		newcount -= nnew;
	    }
	}
	if (oldcount > 0)
	    memcpy(pdata, pdata0, oldcount * sizeof(*pdata));
	pdata += newcount;
    }

    /* set up element data */
    for (i = 0; i < g->nelem; i++) {
	newcount = (hp->elem_index[i + 1] - hp->elem_index[i]) * u->dim;
	if (newcount == 0)
	    continue;
	newtype = hp->info->types[hp->elem_order[i]];
	oldcount = DofGetDataCountOnElement(u, i);
	if (oldcount == 0) {
	    pdata += newcount;
	    continue;
	}
	if (!DofIsHP(u) || u->hp->info != hp->info ||
	    oldcount > newcount || (has_userfunc && oldcount != newcount)) {
	    newtype->InitFunc(utmp, g->elems[i], ELEMENT, 0,
				func0, func1, NULL, pdata, NULL);
	    pdata += newcount;
	    continue;
	}
	oldtype = u->hp->info->types[u->hp->elem_order[i]];
	pdata0 = DofElementData(u, i);
	if (newtype->base_type == NULL) {
	    if (newtype != oldtype && hp->info->types == DOF_HCn) {
		/* copy face-based DOF */
		int nnew = DOF_HBn[newtype->order]->np_face * u->dim;
		int nold = DOF_HBn[oldtype->order]->np_face * u->dim;
		assert(oldtype->order >= 3 && newtype->order >= 3);
		for (ii = 0; ii < 4; ii++) {
		    memcpy(pdata, pdata0, nold * sizeof(*pdata));
		    pdata0 += nold;
		    pdata += nnew;
		    oldcount -= nold;
		    newcount -= nnew;
		}
	    }
	    if (oldcount > 0)
		memcpy(pdata, pdata0, oldcount * sizeof(*pdata));
	    pdata += newcount;
	    continue;
	}
	/* Discontinuous Galerkin bases */
	newtype = newtype->base_type;
	oldtype = oldtype->base_type;
	assert(oldtype != NULL);
	/* Note: can only copy data for Lagrangian bases of same order
	 *	 or hierarchical bases in the DG case */
	if (oldtype != newtype &&
	    (oldtype->points != NULL || newtype->points != NULL)) {
	    newtype->InitFunc(utmp, g->elems[i], ELEMENT, 0,
				func0, func1, NULL, pdata, NULL);
	    pdata += newcount;
	    continue;
	}
	/* simply copy data */
	/* vertex DOF */
	oldcount = u->dim * oldtype->np_vert * NVert;
	newcount = u->dim * newtype->np_vert * NVert;
	assert(oldcount == newcount);
	if (oldcount > 0) {
	    memcpy(pdata, pdata0, oldcount * sizeof(*pdata));
	    pdata0 += oldcount;
	    pdata += newcount;
	}
	/* edge DOF */
	oldcount = u->dim * oldtype->np_edge;
	newcount = u->dim * newtype->np_edge;
	if (oldcount > 0) {
	    for (k = 0; k < NEdge; k++) {
		memcpy(pdata, pdata0, oldcount * sizeof(*pdata));
		pdata0 += oldcount;
		pdata += newcount;
	    }
	}
	else {
	    pdata += newcount * NEdge;
	}
	/* face DOF */
	oldcount = u->dim * oldtype->np_face;
	newcount = u->dim * newtype->np_face;
	if (oldcount > 0) {
	    for (k = 0; k < NFace; k++) {
		memcpy(pdata, pdata0, oldcount * sizeof(*pdata));
		pdata0 += oldcount;
		pdata += newcount;
	    }
	}
	else {
	    pdata += newcount * NFace;
	}
	/* element DOF */
	oldcount = u->dim * oldtype->np_elem;
	newcount = u->dim * newtype->np_elem;
	if (oldcount > 0) {
	    memcpy(pdata, pdata0, oldcount * sizeof(*pdata));
	    pdata0 += oldcount;
	}
	pdata += newcount;
    }

    phgDofFreeVEFMap(&vef);
    utmp->data = NULL;
    phgDofFree(&utmp);

skip:
    /* update members of u */
    phgFree(u->data);
    u->data = u->data_vert = data;
    u->data_edge = u->data_vert + g->nvert * u->count_vert;
    u->data_face = u->data_edge + hp->edge_index[g->nedge] * u->dim;
    u->data_elem = u->data_face + hp->face_index[g->nface] * u->dim;

    phgDofClearCache(NULL, u, NULL, NULL, FALSE);
}

void
phgHPAttachDof(HP_TYPE *hp, DOF *u, BOOLEAN interpolate)
{
    size_t len;

    if (u->hp == hp)
	return;

    if ((!DofIsHP(u) && hp->info != u->type->hp_info) ||
	( DofIsHP(u) && hp->info != u->hp->info))
	phgError(1, "%s: can't use \"%s\" for DOF \"%s\".\n",
		    hp->info->name, u->name);

    update_dof(u, hp, interpolate);
    phgHPFree(&u->hp);
    u->hp = hp;

    /* attach u to hp->dof */
    len = 0;
    if (hp->dof != NULL) {
	DOF **p;
	for (p = hp->dof; *p != NULL; p++, len++);
    }
    hp->dof = phgRealloc(hp->dof, (len + 2) * sizeof(*(hp->dof)));
    hp->dof[len] = u;
    hp->dof[len + 1] = NULL;

    hp->refcount++;

    if (!DofIsHP(u))
	phgDofUnrefId(u->type);

    /* reset count_{vert,edge,face,elem} using max order */
    u->type = hp->info->types[hp->max_order];
    u->count_vert = u->type->np_vert * u->dim;
    u->count_edge = u->type->np_edge * u->dim;
    u->count_face = u->type->np_face * u->dim;
    u->count_elem = u->type->np_elem * u->dim;
    u->type = NULL;	/* accessing u->type will produce segfault */

    phgDofClearCache(NULL, u, NULL, NULL, FALSE);
}

void
phgHPSetup(HP_TYPE *hp, BOOLEAN interpolate)
/* Note: requested order on element e is stored in e->hp_order */
{
    HP_TYPE *hp_old;
    GRID *g = hp->g;
    int i;
    SIMPLEX *e;

    if (hp == NULL)
	return;

    if (hp->derived)
	phgError(1, "shouldn't change orders of a derived HP_TYPE "
		    "(created at %s:%d)\n", hp->srcfile, hp->srcline);

    phgDebug((2, "setting up hierarchical orders.\n"));
    phgDebug((2, "g->nelem = %d.\n", g->nelem));

    hp_old = phgHPDup__(hp, TRUE);

    /* set element orders */
    hp->elem_order = phgCalloc(g->nelem, sizeof(*hp->elem_order));
    ForAllElements(g, e) {
	if (e->hp_order < hp->info->min_order)
	    hp->elem_order[e->index] = hp->info->min_order;
	else if (e->hp_order > hp->info->max_order)
	    hp->elem_order[e->index] = hp->info->max_order;
	else
	    hp->elem_order[e->index] = e->hp_order;
    }

    /* compute edge and face order */
    phgHPSetupOrders(hp);

    /* update DOFs using the HP_TYPE */
    if (hp->dof != NULL) {
	/* note: hp->dof might be reallocated in update_dof() by phgDofNew_ */
	for (i = 0; hp->dof[i] != NULL; i++) {
	    DOF *u = hp->dof[i];
	    u->hp = hp_old;
	    update_dof(u, hp, interpolate);
	    u->hp = hp;
	}
    }

    phgHPFreeBuffers__(hp_old);
    phgFree(hp_old);

    /* This will detach hp->derivative from hp */
    if (hp->derivative != NULL)
	phgHPFree(&hp->derivative);

    /* This will detach hp->dg from hp */
    if (hp->dg != NULL)
	phgHPFree(&hp->dg);

    phgDofClearCache(NULL, NULL, NULL, hp, FALSE);
}

HP_TYPE *
phgHPGetDerivative_(HP_TYPE *old_hp, const char *srcfile, int srcline)
/* returns a new HP_TYPE for the derivatives of 'old_hp' */
{
    GRID *g;
    HP_TYPE *hp;
    INT i;
    int order, min_order, max_order;

    if ((hp = old_hp->derivative) != NULL) {
	hp->refcount++;
	return hp;
    }

    hp = phgHPNew_(g = old_hp->g, HP_DG, -1, srcfile, srcline);
    old_hp->derivative = hp;
    hp->refcount++;	/* referenced by old_hp->derivative */

    min_order = hp->info->min_order;
    max_order = hp->info->max_order;

    for (i = 0; i < g->nedge; i++) {
	if (g->types_edge[i] == UNREFERENCED)
	    continue;
	order = ((int)old_hp->info->types[old_hp->edge_order[i]]->order) - 1;
	if (order < min_order)
	    order = min_order;
	hp->edge_order[i] = order;
    }

    for (i = 0; i < g->nface; i++) {
	if (g->types_face[i] == UNREFERENCED)
	    continue;
	order = ((int)old_hp->info->types[old_hp->face_order[i]]->order) - 1;
	if (order < min_order)
	    order = min_order;
	hp->face_order[i] = order;
    }

    for (i = 0; i < g->nelem; i++) {
	if (g->types_elem[i] == UNREFERENCED)
	    continue;
	order = ((int)old_hp->info->types[old_hp->elem_order[i]]->order) - 1;
	if (order < min_order)
	    order = min_order;
	hp->elem_order[i] = order;
    }

    order = ((int)old_hp->info->types[old_hp->min_order]->order) - 1;
    if (order < min_order)
	order = min_order;
    hp->min_order = order;

    order = ((int)old_hp->info->types[old_hp->max_order]->order) - 1;
    if (order < min_order)
	order = min_order;
    hp->max_order = order;

    phgHPSetupDataPointers(hp);

    return hp;
}

HP_TYPE *
phgHPGetDG_(HP_TYPE *old_hp, const char *srcfile, int srcline)
/* returns a new DG type HP_TYPE of the same order as 'old_hp' */
{
    GRID *g;
    HP_TYPE *hp;
    INT i;
    int order, min_order, max_order;

    if ((hp = old_hp->dg) != NULL) {
	hp->refcount++;
	return hp;
    }

    hp = phgHPNew_(g = old_hp->g, HP_DG, -1, srcfile, srcline);
    old_hp->dg = hp;
    hp->refcount++;	/* referenced by old_hp->dg */

    memset(hp->edge_order, 0, g->nedge * sizeof(*hp->edge_order));
    memset(hp->face_order, 0, g->nface * sizeof(*hp->face_order));

    min_order = HP_DG->max_order;
    max_order = HP_DG->min_order;
    for (i = 0; i < g->nelem; i++) {
	if (g->types_elem[i] == UNREFERENCED)
	    continue;
	order = old_hp->info->types[old_hp->elem_order[i]]->order;
	if (order < HP_DG->min_order)
	    order = HP_DG->min_order;
	else if (order > HP_DG->max_order)
	    order = HP_DG->max_order;

	hp->elem_order[i] = order;

	if (min_order > order)
	    min_order = order;
	if (max_order < order)
	    max_order = order;
    }

    hp->min_order = min_order;
    hp->max_order = max_order;

    phgHPSetupDataPointers(hp);

    return hp;
}

HP_TYPE *
phgHPDup__(HP_TYPE *hp_old, BOOLEAN free_orig_data)
/* Note: this function creates an 'invisible' HP_TYPE which is not registered
 * to GRID, should only be used as scratch variable, and be freed using
 * phgHPFreeBuffers__().
 *
 * If 'free_orig_data' then the xxxx_order and xxxx_index pointers in the
 * original HP_TYPE will be set to NULL. */
{
    GRID *g = hp_old->g;
    HP_TYPE *hp;

    hp = phgAlloc(sizeof(*hp));
    memcpy(hp, hp_old, sizeof(*hp));
    hp->dof = NULL;
    hp->refcount = 0;
    hp->srcfile = NULL;
    hp->derivative = NULL;
    hp->dg = NULL;

    if (free_orig_data) {
	hp_old->edge_order = NULL;
	hp_old->face_order = NULL;
	hp_old->elem_order = NULL;
	hp_old->edge_index = NULL;
	hp_old->face_index = NULL;
	hp_old->elem_index = NULL;
    }
    else {
	hp->edge_order = phgAlloc(g->nedge * sizeof(*hp->edge_order));
	memcpy(hp->edge_order, hp_old->edge_order,
		g->nedge * sizeof(*hp->edge_order));

	hp->face_order = phgAlloc(g->nface * sizeof(*hp->face_order));
	memcpy(hp->face_order, hp_old->face_order,
		g->nface * sizeof(*hp->face_order));

	hp->elem_order = phgAlloc(g->nelem * sizeof(*hp->elem_order));
	memcpy(hp->elem_order, hp_old->elem_order,
		g->nelem * sizeof(*hp->elem_order));

	hp->edge_index = phgAlloc(g->nedge * sizeof(*hp->edge_index));
	memcpy(hp->edge_index, hp_old->edge_index,
		g->nedge * sizeof(*hp->edge_index));

	hp->face_index = phgAlloc(g->nface * sizeof(*hp->face_index));
	memcpy(hp->face_index, hp_old->face_index,
		g->nface * sizeof(*hp->face_index));

	hp->elem_index = phgAlloc(g->nelem * sizeof(*hp->elem_index));
	memcpy(hp->elem_index, hp_old->elem_index,
		g->nelem * sizeof(*hp->elem_index));
    }

    return hp;
}

void
phgHPFreeBuffers__(HP_TYPE *hp)
/* frees all memory blocks allocated by 'hp' */
{
    phgFree(hp->srcfile);
    hp->srcfile = NULL;
    phgFree(hp->edge_order);
    hp->edge_order = NULL;
    phgFree(hp->face_order);
    hp->face_order = NULL;
    phgFree(hp->elem_order);
    hp->elem_order = NULL;
    phgFree(hp->edge_index);
    hp->edge_index = NULL;
    phgFree(hp->face_index);
    hp->face_index = NULL;
    phgFree(hp->elem_index);
    hp->elem_index = NULL;
}

void
phgHPFree(HP_TYPE **hp_ptr)
{
    HP_TYPE *hp;
    HP_TYPE **p, **p0 = NULL;

    if (hp_ptr == NULL || (hp = *hp_ptr) == NULL)
	return;

    if (hp->refcount > 0) {
	--hp->refcount;
	return;
    }

    *hp_ptr = NULL;

    /* detach hp from g->hp */
    for (p = hp->g->hp; p != NULL && *p != NULL; p++) {
	if (*p == hp)
	    p0 = p;
    }
    if (p == NULL || p0 == NULL) {
	phgError(1, "%s:%d: unexpected error(grid freed before HP_TYPE?)\n",
			__FILE__, __LINE__);
    }
    assert(p > p0);
    memmove(p0, p0 + 1, sizeof(*p0) * (p - p0));

    /* free hp->dof */
    if (hp->dof != NULL) {
	if (*hp->dof != NULL)
	    phgWarning("%s: possibly unfreed DOFs found.\n", __func__);
	phgFree(hp->dof);
    }

    phgDofClearCache(NULL, NULL, NULL, hp, TRUE);

    phgHPFree(&hp->derivative);
    phgHPFree(&hp->dg);
    phgHPFreeBuffers__(hp);
    phgFree(hp);
}
